from typing import Any, Literal
from unittest.mock import AsyncMock

import pytest
import respx
from openai import AsyncStream
from openai.types.chat import ChatCompletion, ChatCompletionChunk
from openai.types.chat.chat_completion_chunk import Choice, ChoiceDelta
from openai.types.completion_usage import (
    CompletionTokensDetails,
    CompletionUsage,
    PromptTokensDetails,
)

from azure_switchboard import (
    Deployment,
    Model,
    Switchboard,
)
from azure_switchboard.deployment import DeploymentState


async def collect_chunks(
    stream: AsyncStream[ChatCompletionChunk],
) -> tuple[list[ChatCompletionChunk], str]:
    """Collect all chunks from a stream and return the chunks and assembled content."""
    received_chunks = []
    content = ""
    async for chunk in stream:
        received_chunks.append(chunk)
        if chunk.choices and chunk.choices[0].delta.content:
            content += chunk.choices[0].delta.content
    return received_chunks, content


def openai_config(name: str = "openai") -> Deployment:
    """Create an OpenAI deployment config for testing."""
    return Deployment(
        name=name,
        api_key="test",
        models=[Model(name="gpt-4o-mini"), Model(name="gpt-4o")],
    )


def azure_config(name: str) -> Deployment:
    return Deployment(
        name=name,
        base_url=f"https://{name}.openai.azure.com/openai/v1/",
        api_key=name,
        models=[
            Model(name="gpt-4o-mini", tpm=10000, rpm=60),
            Model(name="gpt-4o", tpm=10000, rpm=60),
        ],
    )


def chat_completion_mock():
    """Basic mock that replicates openai client chat completion behavior."""

    async def _stream(items: list):
        for item in items:
            yield item

    def side_effect(*args, **kwargs):
        if "stream" in kwargs:
            return _stream(COMPLETION_STREAM_CHUNKS)
        return COMPLETION_RESPONSE

    return AsyncMock(side_effect=side_effect)


@pytest.fixture(autouse=True)
def mock_client(request: pytest.FixtureRequest):
    with respx.mock(assert_all_called=False) as respx_mock:
        if request.node.get_closest_marker("mock_models"):
            # Azure deployments use /openai/v1/chat/completions
            respx_mock.route(
                name="azure", method="POST", path="/openai/v1/chat/completions"
            ).respond(json=COMPLETION_RESPONSE_JSON)
            # OpenAI deployments use /v1/chat/completions
            respx_mock.route(
                name="openai", method="POST", path="/v1/chat/completions"
            ).respond(json=COMPLETION_RESPONSE_JSON)

        yield respx_mock


@pytest.fixture
def model():
    return Model(name="gpt-4o-mini", tpm=1000, rpm=6)


@pytest.fixture
def deployment():
    return DeploymentState(azure_config("test1"))


@pytest.fixture
async def switchboard():
    deployments = [
        azure_config("test1"),
        azure_config("test2"),
        azure_config("test3"),
    ]
    async with Switchboard(deployments=deployments, ratelimit_window=0) as sb:
        yield sb


COMPLETION_PARAMS: dict[Literal["model", "messages"], Any] = {
    "model": "gpt-4o-mini",
    "messages": [{"role": "user", "content": "Hello, world!"}],
}

COMPLETION_STREAM_CHUNKS = [
    ChatCompletionChunk(
        id="test_chunk_1",
        choices=[
            Choice(
                delta=ChoiceDelta(content="Hello", role="assistant"),
                finish_reason=None,
                index=0,
            )
        ],
        created=1234567890,
        model="gpt-4o-mini",
        object="chat.completion.chunk",
        usage=None,
    ),
    ChatCompletionChunk(
        id="test_chunk_2",
        choices=[
            Choice(
                delta=ChoiceDelta(content=", "),
                finish_reason=None,
                index=0,
            )
        ],
        created=1234567890,
        model="gpt-4o-mini",
        object="chat.completion.chunk",
        usage=None,
    ),
    ChatCompletionChunk(
        id="test_chunk_3",
        choices=[
            Choice(
                delta=ChoiceDelta(content="world!"),
                finish_reason=None,
                index=0,
            )
        ],
        created=1234567890,
        model="gpt-4o-mini",
        object="chat.completion.chunk",
        usage=None,
    ),
    ChatCompletionChunk(
        id="test_chunk_4",
        choices=[
            Choice(
                delta=ChoiceDelta(),
                finish_reason="stop",
                index=0,
            )
        ],
        created=1234567890,
        model="gpt-4o-mini",
        object="chat.completion.chunk",
        usage=CompletionUsage(
            completion_tokens=5,
            prompt_tokens=15,
            total_tokens=20,
            completion_tokens_details=CompletionTokensDetails(reasoning_tokens=5),
            prompt_tokens_details=PromptTokensDetails(cached_tokens=15),
        ),
    ),
]


COMPLETION_RESPONSE_JSON = {
    "choices": [
        {
            "finish_reason": "stop",
            "index": 0,
            "logprobs": None,
            "message": {
                "content": "Hello! How can I assist you today?",
                "refusal": None,
                "role": "assistant",
            },
        }
    ],
    "created": 1741124380,
    "id": "chatcmpl-test",
    "model": "gpt-4o-mini",
    "object": "chat.completion",
    "service_tier": "default",
    "system_fingerprint": "fp_06737a9306",
    "usage": {
        "completion_tokens": 10,
        "completion_tokens_details": {
            "accepted_prediction_tokens": 0,
            "audio_tokens": 0,
            "reasoning_tokens": 5,
            "rejected_prediction_tokens": 0,
        },
        "prompt_tokens": 10,
        "prompt_tokens_details": {"audio_tokens": 0, "cached_tokens": 8},
        "total_tokens": 20,
    },
}

COMPLETION_RESPONSE = ChatCompletion.model_validate(COMPLETION_RESPONSE_JSON)



import random

import pytest

from azure_switchboard import Switchboard

from .conftest import COMPLETION_PARAMS


class TestAdvanced:
    """Test advanced features of the switchboard."""

    @pytest.mark.mock_models("gpt-4o-mini")
    async def test_round_robin_selector(self, switchboard: Switchboard):
        switchboard.selector = lambda _, options: random.choice(options)

        for _ in range(100):
            await switchboard.create(**COMPLETION_PARAMS)

        # print(switchboard.get_usage())



import azure_switchboard
from loguru import logger as _logger

from azure_switchboard import Deployment, Model, Switchboard


class TestInit:
    def test_public_exports_do_not_include_logging_helpers(self):
        assert "enable_logging" not in azure_switchboard.__all__
        assert "disable_logging" not in azure_switchboard.__all__

    def test_logging_activation_controls_switchboard_logs(self):
        records: list[dict] = []
        sink_id = _logger.add(lambda m: records.append(m.record))
        switchboard = Switchboard(
            deployments=[
                Deployment(
                    name="mini-only",
                    base_url="https://mini-only.openai.azure.com/openai/v1/",
                    api_key="mini-only",
                    models=[Model(name="gpt-4o-mini", tpm=1000, rpm=6)],
                ),
                Deployment(
                    name="full-only",
                    base_url="https://full-only.openai.azure.com/openai/v1/",
                    api_key="full-only",
                    models=[Model(name="gpt-4o", tpm=1000, rpm=6)],
                ),
            ],
            ratelimit_window=0,
        )
        try:
            _logger.disable("azure_switchboard")
            switchboard.sessions["test"] = switchboard.deployments["mini-only"]
            _ = switchboard.select_deployment(session_id="test", model="gpt-4o")
            assert not records

            _logger.enable("azure_switchboard")
            switchboard.sessions["test"] = switchboard.deployments["mini-only"]
            _ = switchboard.select_deployment(session_id="test", model="gpt-4o")
            assert any(
                "is unhealthy on mini-only, falling back to selection" in r["message"]
                for r in records
            )
        finally:
            _logger.remove(sink_id)
            _logger.disable("azure_switchboard")



from __future__ import annotations

from collections.abc import AsyncIterator
from typing import Literal, cast, overload

import wrapt
from loguru import logger
from openai import AsyncOpenAI, AsyncStream, RateLimitError
from openai.types.chat import ChatCompletion, ChatCompletionChunk
from openai.types.completion_usage import CompletionUsage
from opentelemetry import trace
from pydantic import BaseModel, Field

from azure_switchboard.model import UtilStats

from .exceptions import SwitchboardError
from .model import Model


class Deployment(BaseModel, arbitrary_types_allowed=True):
    """Configuration for an Azure OpenAI or OpenAI deployment."""

    name: str
    base_url: str | None = None
    api_key: str | None = None
    timeout: float = 600.0
    models: list[Model] = Field(
        default_factory=lambda: [
            Model(name="gpt-4o"),
            Model(name="gpt-4o-mini"),
            Model(name="gpt-4.1"),
            Model(name="gpt-4.1-mini"),
            Model(name="gpt-4.1-nano"),
            Model(name="gpt-5"),
            Model(name="gpt-5-mini"),
            Model(name="gpt-5-nano"),
            Model(name="gpt-5-chat"),
            Model(name="gpt-5.1"),
            Model(name="gpt-5.1-mini"),
            Model(name="gpt-5.1-nano"),
            Model(name="gpt-5.1-chat"),  # OpenAI calls this gpt-5.1-chat-latest
            Model(name="gpt-5.2"),
            Model(name="gpt-5.2-mini"),
            Model(name="gpt-5.2-nano"),
            Model(name="gpt-5.2-chat"),  # OpenAI calls this gpt-5.2-chat-latest
        ]
    )
    client: AsyncOpenAI | None = None

    def get_client(self) -> AsyncOpenAI:
        return AsyncOpenAI(
            api_key=self.api_key,
            base_url=self.base_url,
            timeout=self.timeout,
        )


class DeploymentState:
    """Runtime state of a deployment"""

    def __init__(
        self,
        config: Deployment,
    ) -> None:
        self.config = config
        self.client = config.get_client()
        self.models = {m.name: m for m in config.models}

    def __repr__(self) -> str:
        elems = ", ".join(map(str, self.models.values()))
        return f"Deployment<{self.config.name}>([{elems}])"

    @property
    def name(self) -> str:
        return self.config.name

    def reset_usage(self) -> None:
        for model in self.models.values():
            model.reset_usage()

    def stats(self) -> dict[str, UtilStats]:
        return {name: model.stats() for name, model in self.models.items()}

    def is_healthy(self, model: str) -> bool:
        return self.model(model).is_healthy() if model in self.models else False

    def util(self, model: str) -> float:
        return self.model(model).util if model in self.models else 0.0

    def model(self, name: str) -> Model:
        return self.models[name]

    @overload
    async def create(
        self, *, model: str, stream: Literal[True], **kwargs
    ) -> AsyncStream[ChatCompletionChunk]: ...

    @overload
    async def create(self, *, model: str, **kwargs) -> ChatCompletion: ...

    async def create(
        self,
        *,
        model: str,
        stream: bool = False,
        **kwargs,
    ) -> ChatCompletion | AsyncStream[ChatCompletionChunk]:
        """
        Send a chat completion request to this client.
        Tracks usage metrics for load balancing.
        """

        if model not in self.models:
            raise SwitchboardError(f"{model} not configured for deployment")

        # add input token estimate before we send the request so utilization is
        # kept up to date for other requests that might be executing concurrently.
        _preflight_estimate = self._estimate_token_usage(kwargs)
        self.models[model].spend_tokens(_preflight_estimate)
        self.models[model].spend_request()

        kwargs["timeout"] = kwargs.get("timeout", self.config.timeout)
        try:
            if stream:
                logger.trace("Creating streaming completion")
                response_stream = await self.client.chat.completions.create(
                    model=model,
                    stream=True,
                    stream_options=kwargs.pop(
                        "stream_options", {"include_usage": True}
                    ),
                    **kwargs,
                )

                # streaming util gets updated inside _AsyncStreamWrapper
                return _AsyncStreamWrapper(
                    stream=response_stream,
                    deployment=self,
                    model=self.models[model],
                    offset=_preflight_estimate,
                )

            else:
                logger.trace("Creating chat completion")
                response = await self.client.chat.completions.create(
                    model=model, **kwargs
                )
                response = cast(ChatCompletion, response)

                if response.usage:
                    self.models[model].spend_tokens(
                        # dont double-count our preflight estimate
                        response.usage.total_tokens - _preflight_estimate
                    )
                    self._set_span_attributes(response.usage)

                return response
        except RateLimitError as e:
            logger.warning("Hit rate limits")
            self.models[model].mark_down()
            raise SwitchboardError(
                "Rate limit exceeded in deployment chat completion"
            ) from e
        except Exception as e:
            logger.exception("Marking down model for chat completion error")
            self.models[model].mark_down()
            raise RuntimeError("Error in deployment chat completion") from e

    def _estimate_token_usage(self, kwargs: dict) -> int:
        # loose estimate of token cost. were only considering
        # input tokens for now, we can add output estimates as well later.
        # openai says roughly 4 characters per token, so sum len of messages
        # and divide by 4.
        t_input = sum(len(m.get("content", "")) for m in kwargs.get("messages", []))
        # t_output = kwargs.get("max_tokens", 500)
        return t_input // 4

    def _set_span_attributes(self, usage: CompletionUsage) -> None:
        span = trace.get_current_span()
        if prompt := usage.prompt_tokens_details:
            if prompt.cached_tokens:
                span.set_attribute("gen_ai.usage.cached_tokens", prompt.cached_tokens)
        if completion := usage.completion_tokens_details:
            if completion.reasoning_tokens:
                span.set_attribute(
                    "gen_ai.usage.reasoning_tokens", completion.reasoning_tokens
                )


class _AsyncStreamWrapper(wrapt.ObjectProxy):
    """Wrap an openai.AsyncStream to track usage"""

    def __init__(
        self,
        stream: AsyncStream[ChatCompletionChunk],
        deployment: DeploymentState,
        model: Model,
        offset: int = 0,
    ):
        super().__init__(stream)
        self._self_deployment: DeploymentState = deployment
        self._self_model: Model = model
        self._self_offset: int = offset
        # Stream chunks are consumed after Switchboard.create() returns, so the
        # contextualized logger scope from create() is no longer active here.
        # Keep a bound logger on the wrapper to retain deployment/model context
        # for mid-stream error logs.
        self._self_logger = logger.bind(
            deployment=deployment.name,
            model=model.name,
        )

    async def __aiter__(self) -> AsyncIterator[ChatCompletionChunk]:
        try:
            async for chunk in self.__wrapped__:
                chunk = cast(ChatCompletionChunk, chunk)
                # only the last chunk contains the usage info
                if chunk.usage:
                    self._self_model.spend_tokens(
                        # dont double-count our preflight estimate
                        chunk.usage.total_tokens - self._self_offset
                    )
                    self._self_deployment._set_span_attributes(chunk.usage)

                yield chunk
        except Exception as e:
            self._self_logger.exception("Marking down model for wrapped stream error")
            self._self_model.mark_down()
            if isinstance(e, RateLimitError):
                raise SwitchboardError("Rate limit exceeded in wrapped stream") from e
            raise RuntimeError("Error in wrapped stream") from e



from __future__ import annotations

import asyncio
import random
from collections import OrderedDict
from typing import Callable, Literal, Sequence, overload

from loguru import logger
from openai import AsyncStream
from openai.types.chat import ChatCompletion, ChatCompletionChunk
from opentelemetry import metrics
from tenacity import (
    AsyncRetrying,
    retry_if_not_exception_type,
    stop_after_attempt,
)

from azure_switchboard.model import UtilStats

from .deployment import (
    Deployment,
    DeploymentState,
)
from .exceptions import SwitchboardError

meter = metrics.get_meter("azure_switchboard.switchboard")
deployment_util = meter.create_gauge(
    name="switchboard.deployment.model.utilization",
    description="Utilization of a model on a deployment",
    unit="%",
)
healthy_deployments_gauge = meter.create_gauge(
    name="healthy_deployments_count",
    description="Number of healthy deployments available for a model",
    unit="1",
)
deployment_failures_counter = meter.create_counter(
    name="deployment_failures",
    description="Number of deployment failures",
    unit="1",
)
request_counter = meter.create_counter(
    name="requests",
    description="Number of requests sent through the switchboard",
    unit="1",
)


def two_random_choices(model: str, options: list[DeploymentState]) -> DeploymentState:
    """Power of two random choices algorithm.

    Randomly select 2 deployments and return the one
    with lower util for the given model.
    """
    selected = random.sample(options, min(2, len(options)))
    return min(selected, key=lambda d: d.util(model))


DEFAULT_FAILOVER_POLICY = AsyncRetrying(
    stop=stop_after_attempt(2),
    retry=retry_if_not_exception_type(SwitchboardError),
    reraise=True,
)


class Switchboard:
    def __init__(
        self,
        deployments: Sequence[Deployment],
        selector: Callable[
            [str, list[DeploymentState]], DeploymentState
        ] = two_random_choices,
        failover_policy: AsyncRetrying = DEFAULT_FAILOVER_POLICY,
        ratelimit_window: float = 60.0,
        max_sessions: int = 1024,
    ) -> None:
        if not deployments:
            raise SwitchboardError("No deployments provided")

        self.deployments: dict[str, DeploymentState] = {}
        for deployment in deployments:
            if deployment.name in self.deployments:
                raise SwitchboardError(f"Duplicate deployment name: {deployment.name}")
            self.deployments[deployment.name] = DeploymentState(deployment)

        self.selector = selector
        self.failover_policy = failover_policy

        # LRUDict to expire old sessions
        self.sessions = _LRUDict(max_size=max_sessions)
        self.ratelimit_reset_task: asyncio.Task | None = None

        # reset usage every N seconds
        self.ratelimit_window = ratelimit_window

    async def __aenter__(self) -> Switchboard:
        self.start()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback) -> None:
        await self.stop()

    def start(self) -> None:
        async def periodic_reset():
            if not self.ratelimit_window:
                return

            while True:
                await asyncio.sleep(self.ratelimit_window)
                self.reset_usage()

        self.ratelimit_reset_task = asyncio.create_task(periodic_reset())

    async def stop(self) -> None:
        if self.ratelimit_reset_task:
            try:
                self.ratelimit_reset_task.cancel()
                await self.ratelimit_reset_task
            except asyncio.CancelledError:
                pass

    def reset_usage(self) -> None:
        for deployment in self.deployments.values():
            deployment.reset_usage()

    def stats(self) -> dict[str, dict[str, UtilStats]]:
        return {
            name: deployment.stats() for name, deployment in self.deployments.items()
        }

    def select_deployment(
        self, *, model: str, session_id: str | None = None
    ) -> DeploymentState:
        """
        Select a deployment using the power of two random choices algorithm.
        If session_id is provided, try to use that specific deployment first.
        """
        # Handle session-based routing first
        if session_id and session_id in self.sessions:
            deployment = self.sessions[session_id]
            if deployment.is_healthy(model):
                return deployment

            m = deployment.models.get(model)
            logger.bind(util=vars(m.stats()) if m else None).warning(
                f"{model} is unhealthy on {deployment.name}, falling back to selection"
            )

        # Get eligible deployments for the requested model
        eligible_deployments = [
            d for d in self.deployments.values() if d.is_healthy(model)
        ]

        if not eligible_deployments:
            raise SwitchboardError(f"No eligible deployments available for {model}")

        # Record healthy deployments count metric
        healthy_deployments_gauge.set(len(eligible_deployments), {"model": model})

        if len(eligible_deployments) == 1:
            deployment = eligible_deployments[0]
        else:
            deployment = self.selector(model, eligible_deployments)

        logger.trace(f"Selected deployment: {deployment.name}")

        if session_id:
            self.sessions[session_id] = deployment

        return deployment

    @overload
    async def create(
        self, *, session_id: str | None = None, stream: Literal[True], **kwargs
    ) -> AsyncStream[ChatCompletionChunk]: ...

    @overload
    async def create(
        self, *, session_id: str | None = None, **kwargs
    ) -> ChatCompletion: ...

    async def create(
        self,
        *,
        model: str,
        session_id: str | None = None,
        stream: bool = False,
        **kwargs,
    ) -> ChatCompletion | AsyncStream[ChatCompletionChunk]:  # pyright: ignore[reportReturnType]
        """
        Send a chat completion request to the selected deployment, with automatic failover.
        """
        with logger.contextualize(model=model, session_id=session_id):
            async for attempt in self.failover_policy:
                with attempt:
                    deployment = self.select_deployment(
                        model=model, session_id=session_id
                    )
                    with logger.contextualize(deployment=deployment.name):
                        logger.trace("Sending completion request")
                        response = await deployment.create(
                            model=model, stream=stream, **kwargs
                        )
                    request_counter.add(
                        1, {"model": model, "deployment": deployment.name}
                    )
                    return response

    def __repr__(self) -> str:
        return f"Switchboard({self.deployments})"


# borrowed from https://gist.github.com/davesteele/44793cd0348f59f8fadd49d7799bd306
class _LRUDict(OrderedDict):
    def __init__(self, *args, max_size: int = 1024, **kwargs):
        assert max_size > 0
        self.max_size = max_size

        super().__init__(*args, **kwargs)

    def __setitem__(self, key: str, value: DeploymentState) -> None:
        super().__setitem__(key, value)
        super().move_to_end(key)

        while len(self) > self.max_size:  # pragma: no cover
            oldkey = next(iter(self))
            super().__delitem__(oldkey)

    def __getitem__(self, key: str) -> DeploymentState:
        val = super().__getitem__(key)
        super().move_to_end(key)

        return val



import asyncio


from azure_switchboard import Model


class TestModel:
    """Model functionality tests."""

    async def test_init(self, model: Model):
        assert str(model).startswith("Model<gpt-4o-mini>(util=0.0")

    async def test_util(self, model: Model):
        assert model.is_healthy()

        model.tpm_usage = 2000
        assert not model.is_healthy()

        model.tpm_usage = 500
        assert model.is_healthy()

        model.rpm_usage = 10
        assert not model.is_healthy()

        model.rpm_usage = 5
        assert model.is_healthy()

    async def test_markdown(self, model: Model):
        assert model.is_healthy()

        model.mark_down(1)
        assert not model.is_healthy()

        await asyncio.sleep(0.5)
        assert not model.is_healthy()

        await asyncio.sleep(0.5)
        assert model.is_healthy()

        model.mark_down(10)
        assert not model.is_healthy()

        model.mark_up()
        assert model.is_healthy()



from loguru import logger as _logger

from .deployment import Deployment
from .exceptions import SwitchboardError
from .model import Model
from .switchboard import Switchboard

# As a library, do not configure sinks or emit logs by default.
# Applications can opt in explicitly.
_logger.disable("azure_switchboard")


__all__ = [
    "Deployment",
    "Model",
    "SwitchboardError",
    "Switchboard",
]



import asyncio
from unittest.mock import AsyncMock, patch

import pytest
import respx
from httpx import Request, Response, TimeoutException
from loguru import logger as _logger
from openai import RateLimitError

from azure_switchboard import SwitchboardError
from azure_switchboard.deployment import DeploymentState

from .conftest import (
    COMPLETION_PARAMS,
    COMPLETION_RESPONSE,
    COMPLETION_RESPONSE_JSON,
    COMPLETION_STREAM_CHUNKS,
    chat_completion_mock,
    collect_chunks,
)


class TestDeployment:
    """Deployment functionality tests."""

    async def test_init(self, deployment: DeploymentState):
        """Test initialization of Deployment."""
        assert deployment is not None
        assert deployment.client is not None
        assert deployment.models is not None
        assert deployment.model("gpt-4o-mini") is not None
        assert deployment.model("gpt-4o") is not None

    @pytest.mark.mock_models("gpt-4o-mini")
    async def test_completion(
        self, mock_client: respx.MockRouter, deployment: DeploymentState
    ):
        """Test basic chat completion functionality."""

        deployment.client.max_retries = 0

        response = await deployment.create(**COMPLETION_PARAMS)
        assert mock_client.routes["azure"].call_count == 1
        assert response == COMPLETION_RESPONSE

        # Check token usage tracking
        model = deployment.model("gpt-4o-mini")
        usage = model.stats()
        assert usage.tpm.startswith(str(COMPLETION_RESPONSE.usage.total_tokens))  # pyright: ignore[reportOptionalMemberAccess]
        assert usage.rpm.startswith("1")

        # Test exception handling
        mock_client.routes["azure"].side_effect = Exception("test")
        with pytest.raises(RuntimeError):  # (openai.APIConnectionError):
            await deployment.create(**COMPLETION_PARAMS)
        assert mock_client.routes["azure"].call_count == 2

        # account for preflight estimate
        usage = model.stats()
        assert "23/" in usage.tpm
        assert "2/" in usage.rpm

    async def test_streaming(self, deployment: DeploymentState):
        """Test streaming functionality.

        It's annoying to try to mock HTTP streaming responses so we cheat
        a little bit with an AsyncMock.
        """

        deployment.client.max_retries = 0

        with patch.object(
            deployment.client.chat.completions,
            "create",
            side_effect=chat_completion_mock(),
        ) as mock:
            stream = await deployment.create(stream=True, **COMPLETION_PARAMS)
            mock.assert_called_once()

            # verify basic behavior
            received_chunks, content = await collect_chunks(stream)
            assert len(received_chunks) == len(COMPLETION_STREAM_CHUNKS)
            assert content == "Hello, world!"

            # Verify token usage tracking
            usage = deployment.model("gpt-4o-mini").stats()
            assert "20/" in usage.tpm
            assert "1/" in usage.rpm

        # verify exception handling
        with patch.object(
            deployment.client.chat.completions,
            "create",
            side_effect=Exception("test"),
        ) as mock:
            with pytest.raises(RuntimeError):
                stream = await deployment.create(stream=True, **COMPLETION_PARAMS)
                async for _ in stream:
                    pass
            mock.assert_called_once()

            usage = deployment.model("gpt-4o-mini").stats()
            assert "23/" in usage.tpm
            assert "2/" in usage.rpm

        # Test midstream exception handling
        with patch.object(
            deployment.client.chat.completions,
            "create",
            side_effect=chat_completion_mock(),
        ) as mock:
            stream = await deployment.create(stream=True, **COMPLETION_PARAMS)

            records: list[dict] = []
            sink_id = _logger.add(lambda m: records.append(m.record))
            _logger.enable("azure_switchboard")
            try:
                with patch.object(
                    stream._self_model,  # type: ignore[reportAttributeAccessIssue]
                    "spend_tokens",
                    side_effect=Exception("asyncstream error"),
                ):
                    with pytest.raises(RuntimeError, match="Error in wrapped stream"):
                        await collect_chunks(stream)
                    assert mock.call_count == 1
                    assert not deployment.model("gpt-4o-mini").is_healthy()
                    assert any(
                        r["message"] == "Marking down model for wrapped stream error"
                        and r["extra"].get("deployment") == deployment.name
                        and r["extra"].get("model") == "gpt-4o-mini"
                        for r in records
                    )

                deployment.model("gpt-4o-mini").mark_up()
                assert deployment.model("gpt-4o-mini").is_healthy()
            finally:
                _logger.remove(sink_id)
                _logger.disable("azure_switchboard")

        # Test midstream rate limit handling
        rate_limit_error = RateLimitError(
            "rate limited",
            response=Response(
                status_code=429,
                request=Request(
                    "POST", "https://test.openai.azure.com/openai/v1/chat/completions"
                ),
            ),
            body={"error": {"message": "rate limited"}},
        )

        async def rate_limited_stream():
            yield COMPLETION_STREAM_CHUNKS[0]
            raise rate_limit_error

        with patch.object(
            deployment.client.chat.completions,
            "create",
            new=AsyncMock(return_value=rate_limited_stream()),
        ) as mock:
            stream = await deployment.create(stream=True, **COMPLETION_PARAMS)
            with pytest.raises(
                SwitchboardError, match="Rate limit exceeded in wrapped stream"
            ):
                await collect_chunks(stream)
            assert mock.call_count == 1
            assert not deployment.model("gpt-4o-mini").is_healthy()

    async def test_mark_down(self, deployment: DeploymentState):
        """Test model-level cooldown functionality."""

        model = deployment.model("gpt-4o-mini")

        model.mark_down()
        assert not model.is_healthy()

        model.mark_up()
        assert model.is_healthy()

    async def test_valid_model(self, deployment: DeploymentState):
        """Test that an invalid model raises an error."""

        with pytest.raises(SwitchboardError, match="gpt-fake not configured"):
            await deployment.create(model="gpt-fake", messages=[])

    async def test_usage(self, deployment: DeploymentState):
        """Test client-level counters"""

        # Reset and verify initial state
        for model in deployment.models.values():
            assert "tpm='0" in str(model)

        # Test client-level usage
        model = deployment.model("gpt-4o-mini")
        usage = model.stats()
        assert usage.tpm == f"0/{model.tpm_limit}"
        assert usage.rpm == f"0/{model.rpm_limit}"

        # Set and verify values
        model.spend_tokens(100)
        model.spend_request(5)
        usage = model.stats()
        assert usage.tpm == f"100/{model.tpm_limit}"
        assert usage.rpm == f"5/{model.rpm_limit}"

        # Reset and verify again
        deployment.reset_usage()
        usage = model.stats()
        assert usage.tpm == f"0/{model.tpm_limit}"
        assert usage.rpm == f"0/{model.rpm_limit}"
        assert model.last_reset > 0

    async def test_utilization(self, deployment: DeploymentState):
        """Test utilization calculation."""

        model = deployment.model("gpt-4o-mini")

        # Check initial utilization (nonzero due to random splay)
        initial_util = model.util
        assert 0 <= initial_util < 0.02

        # Test token-based utilization
        model.spend_tokens(5000)  # 50% of TPM limit
        util_with_tokens = model.util
        assert 0.5 <= util_with_tokens < 0.52

        # Test request-based utilization
        model.reset_usage()
        model.spend_request(30)  # 50% of RPM limit
        util_with_requests = model.util
        assert 0.5 <= util_with_requests < 0.52

        # Test combined utilization (should take max of the two)
        model.reset_usage()
        model.spend_tokens(6000)  # 60% of TPM
        model.spend_request(30)  # 50% of RPM
        util_with_both = model.util
        assert 0.6 <= util_with_both < 0.62

        # Test unhealthy client
        model.mark_down()
        assert model.util == 1

    @pytest.mark.mock_models("gpt-4o-mini", "gpt-4o")
    async def test_multiple_models(
        self, mock_client: respx.MockRouter, deployment: DeploymentState
    ):
        """Test that multiple models are handled correctly."""

        gpt4o = deployment.models["gpt-4o"]
        gpt4o_mini = deployment.models["gpt-4o-mini"]

        assert gpt4o is not None
        assert gpt4o_mini is not None

        _ = await deployment.create(
            model="gpt-4o", messages=COMPLETION_PARAMS["messages"]
        )
        assert mock_client.routes["azure"].call_count == 1
        assert gpt4o.rpm_usage == 1
        assert gpt4o.tpm_usage > 0
        assert gpt4o_mini.tpm_usage == 0
        assert gpt4o_mini.rpm_usage == 0

        _ = await deployment.create(**COMPLETION_PARAMS)
        assert mock_client.routes["azure"].call_count == 2

        assert gpt4o.rpm_usage == 1
        assert gpt4o.tpm_usage > 0
        assert gpt4o_mini.tpm_usage > 0
        assert gpt4o_mini.rpm_usage == 1

    @pytest.mark.mock_models("gpt-4o-mini")
    async def test_concurrency(
        self, mock_client: respx.MockRouter, deployment: DeploymentState
    ):
        """Test handling of multiple concurrent requests."""

        # Create and run concurrent requests
        num_requests = 10
        tasks = [deployment.create(**COMPLETION_PARAMS) for _ in range(num_requests)]
        responses = await asyncio.gather(*tasks)

        # Verify results
        model = deployment.models["gpt-4o-mini"]
        assert len(responses) == num_requests
        assert all(r == COMPLETION_RESPONSE for r in responses)
        assert mock_client.routes["azure"].call_count == num_requests
        usage = model.stats()
        assert usage.tpm == f"{20 * num_requests}/10000"
        assert usage.rpm == f"{num_requests}/60"

    @pytest.mark.mock_models("gpt-4o-mini")
    async def test_timeout_retry(
        self, mock_client: respx.MockRouter, deployment: DeploymentState
    ):
        """Test timeout retry behavior."""

        # Test successful retry after timeouts
        expected_response = Response(status_code=200, json=COMPLETION_RESPONSE_JSON)
        mock_client.routes["azure"].side_effect = [
            TimeoutException("Timeout 1"),
            TimeoutException("Timeout 2"),
            expected_response,
        ]
        response = await deployment.create(**COMPLETION_PARAMS)
        assert response == COMPLETION_RESPONSE
        assert mock_client.routes["azure"].call_count == 3

        # Test failure after max retries
        mock_client.routes["azure"].reset()
        mock_client.routes["azure"].side_effect = [
            TimeoutException("Timeout 1"),
            TimeoutException("Timeout 2"),
            TimeoutException("Timeout 3"),
        ]

        with pytest.raises(RuntimeError):  # (openai.APITimeoutError):
            await deployment.create(**COMPLETION_PARAMS)
        assert mock_client.routes["azure"].call_count == 3
        assert not deployment.is_healthy("gpt-4o-mini")

    async def test_rate_limit_is_non_retriable(self, deployment: DeploymentState):
        """Rate limit errors should be surfaced as SwitchboardError."""

        deployment.client.max_retries = 0
        rate_limit_error = RateLimitError(
            "rate limited",
            response=Response(
                status_code=429,
                request=Request(
                    "POST", "https://test.openai.azure.com/openai/v1/chat/completions"
                ),
            ),
            body={"error": {"message": "rate limited"}},
        )

        with patch.object(
            deployment.client.chat.completions,
            "create",
            side_effect=rate_limit_error,
        ) as mock:
            with pytest.raises(
                SwitchboardError,
                match="Rate limit exceeded in deployment chat completion",
            ):
                await deployment.create(**COMPLETION_PARAMS)
            assert mock.call_count == 1
            assert not deployment.model("gpt-4o-mini").is_healthy()

    async def test_invalid_model(self, deployment: DeploymentState):
        """Test that an invalid or unconfigured model is not eligible on a deployment."""

        assert not deployment.is_healthy("invalid-model")



#!/usr/bin/env python3
#
# To run this, use:
#   uv run --env-file .env tools/readme_example.py
#
# /// script
# requires-python = ">=3.10"
# dependencies = [
#     "azure-switchboard",
# ]
# ///

import asyncio
import os

from azure_switchboard import Deployment, Model, Switchboard

azure_openai_endpoint = os.getenv("AZURE_OPENAI_ENDPOINT")
azure_openai_api_key = os.getenv("AZURE_OPENAI_API_KEY")
openai_api_key = os.getenv("OPENAI_API_KEY")

deployments = []
if azure_openai_endpoint and azure_openai_api_key:
    # create 3 deployments. reusing the endpoint
    # is fine for the purposes of this demo
    for name in ("east", "west", "south"):
        deployments.append(
            Deployment(
                name=name,
                base_url=f"{azure_openai_endpoint}/openai/v1/",
                api_key=azure_openai_api_key,
                models=[Model(name="gpt-4o-mini")],
            )
        )

if openai_api_key:
    deployments.append(
        Deployment(
            name="openai",
            api_key=openai_api_key,
            models=[Model(name="gpt-4o-mini")],
        )
    )

if not deployments:
    raise RuntimeError(
        "Set AZURE_OPENAI_ENDPOINT/AZURE_OPENAI_API_KEY or OPENAI_API_KEY to run this example."
    )


async def main():
    async with Switchboard(deployments=deployments) as sb:
        print("Basic functionality:")
        await basic_functionality(sb)

        print("Session affinity (should warn):")
        await session_affinity(sb)


async def basic_functionality(switchboard: Switchboard):
    # Make a completion request (non-streaming)
    response = await switchboard.create(
        model="gpt-4o-mini",
        messages=[{"role": "user", "content": "Hello, world!"}],
    )

    print("completion:", response.choices[0].message.content)

    # Make a streaming completion request
    stream = await switchboard.create(
        model="gpt-4o-mini",
        messages=[{"role": "user", "content": "Hello, world!"}],
        stream=True,
    )

    print("streaming: ", end="")
    async for chunk in stream:
        if chunk.choices and chunk.choices[0].delta.content:
            print(chunk.choices[0].delta.content, end="", flush=True)

    print()


async def session_affinity(switchboard: Switchboard):
    session_id = "anything"

    # First message will select a random healthy
    # deployment and associate it with the session_id
    r = await switchboard.create(
        session_id=session_id,
        model="gpt-4o-mini",
        messages=[{"role": "user", "content": "Who won the World Series in 2020?"}],
    )

    d1 = switchboard.select_deployment(model="gpt-4o-mini", session_id=session_id)
    print("deployment 1:", d1)
    print("response 1:", r.choices[0].message.content)

    # Follow-up requests with the same session_id will route to the same deployment
    r2 = await switchboard.create(
        session_id=session_id,
        model="gpt-4o-mini",
        messages=[
            {"role": "user", "content": "Who won the World Series in 2020?"},
            {"role": "assistant", "content": r.choices[0].message.content},
            {"role": "user", "content": "Who did they beat?"},
        ],
    )

    print("response 2:", r2.choices[0].message.content)

    # Simulate a failure by marking down the deployment
    d1.models["gpt-4o-mini"].mark_down()

    # A new deployment will be selected for this session_id
    r3 = await switchboard.create(
        session_id=session_id,
        model="gpt-4o-mini",
        messages=[{"role": "user", "content": "Who won the World Series in 2021?"}],
    )

    d2 = switchboard.select_deployment(model="gpt-4o-mini", session_id=session_id)
    print("deployment 2:", d2)
    print("response 3:", r3.choices[0].message.content)
    assert d2 != d1


if __name__ == "__main__":
    asyncio.run(main())
